<?php

/**
 * EODataClient.php
 *
 * Integrates the OData PHP SDK into Yii for consuming OData services
 *
 * @link http://www.odata.org/
 * @link http://odataphp.codeplex.com/
 *
 * @author Joe Blocher <yii@myticket.at>
 * @copyright 2011 myticket it-solutions gmbh
 * @license New BSD License
 * @category Webservice
 * @package odataclient
 * @version 0.1
 */
class EODataClient extends CComponent {

	/**
	 * The current config
	 * This is the value of the config array with the key 'serviceName'
	 * @see services/config.php
	 *
	 * @var array $_config
	 */
	private $_config;

	/**
	 * The current serviceName
	 * This is name of the service proxy class, generated by the SDK
	 * @see services/config.php
	 *
	 * @var string $_serviceName
	 */
    private $_serviceName;

	/**
	 * The current uses serviceNameAlias
	 * This is the key of the config array
	 * @see services/config.php
	 *
	 * @var string $_serviceNameAlias
	 */
    private $_serviceNameAlias;

	/**
	 * Internal used to check if service files have been included
	 *
	 * @var boolean $_serviceImported
	 */
    private $_serviceImported;

	/**
	 * Internal used for memory caching
	 *
	 * @var array $_services
	 */
    protected static $_services;

	/**
	 * Internal used for memory caching
	 *
	 * @var array $_metadataXmlElement
	 */
    protected static $_metadataXmlElement;

	/**
	 * Internal used for memory caching
	 *
	 * @var array $_entities
	 */
    protected static $_entities;

    /**
     * Initialize from the config file
     * Generate the proxy class file if not exists (see SDK PHPDataSvcUtil.php
     *
     * @param string $serviceNameAlias the key of the array in the config file
     * @param string $configFileName
     * @param mixed $autoGenerateProxyClass
     */
    public function __construct($serviceNameAlias, $configFileName = 'config.php', $autoGenerateProxyClass = true)
    {
        $this->loadServiceConfig($serviceNameAlias, $configFileName);
        // check if proxy class file exists, otherwise generate
        if ($autoGenerateProxyClass) {
            $serviceFile = $this->getServicefile();
            if (!is_file($serviceFile)) {
                ob_start();
                $this->generateProxyClassFile();
                ob_end_clean();
            }
        }
    }

    /**
     * Get the real path to the config directory
     *
     * @return
     */
    protected static function getConfigPath()
    {
        return dirname(__FILE__) . '/../services';
    }

    /**
     * Get the real path to the config file
     *
     * @return
     */
    protected static function getConfigFile($configFileName)
    {
        $configFile = self::getConfigPath() . '/' . $configFileName;

        if (!is_file($configFile))
            throw new CException('Configfile not found: ' . $configFileName);

        return $configFile;
    }

    /**
     * Initialize from the config file
     *
     * @param string $servicesNameAlias
     * @param string $configFileName
     * @return
     */
    protected function loadServiceConfig($servicesNameAlias, $configFileName)
    {
        $config = new CConfiguration($this->getConfigFile($configFileName));

        $serviceConfig = $config->itemAt($servicesNameAlias);

        if (empty($serviceConfig))
            throw new CException('Service not configured: ' . $servicesNameAlias);

        if (empty($serviceConfig['serviceUri']))
            throw new CException('serviceUri not configured for service: ' . $servicesNameAlias);

        $this->_config = $serviceConfig;
        $this->_serviceNameAlias = $servicesNameAlias;
        $this->_serviceName = isset($serviceConfig['serviceName']) ? $serviceConfig['serviceName'] : $servicesNameAlias;
    }

    /**
     * Get the ServiceName (= the class name / file name of the service proxy class)
     *
     * @return string
     */
    public function getServiceName()
    {
        return $this->_serviceName;
    }

    /**
     * Get the ServiceNameAlias
     *
     * @return string
     */
    public function getServiceNameAlias()
    {
        return $this->_serviceNameAlias;
    }

    /**
     * Create an instance of a configured service (NorthwindEntities ...)
     *
     *   This is equivalent:
     *
     *   1. OData PHP SDK
     *   $proxy = new NorthwindEntities();
     *
     *   2. Yii extension odataclient:
     *
     *   $client = new EODataClient('northwind'); //the service alias from the config file
     *   $proxy = $client->service;
     *
     * @return object
     */
    public function getService()
    {
        if (!isset(self::$_services))
            self::$_services = array();

        $id = $this->getServiceId();

        if (!isset(self::$_services[$id])) {
            $this->importService();
            $className = $this->_serviceName;
            self::$_services[$id] = new $className;
        }

        return self::$_services[$id];
    }

    /**
     * Create an instance of a entitySet
     * This is a object from the service proxy class (Customers, Categories ...)
     * This method must be called when using the ODataClientDataProvider
     *
     * @see ODataController.actionServices

  This is equivalent:

  1. OData PHP SDK

  $proxy = new NorthwindEntities();
  $query = $proxy->Customers()->filter("Country eq 'USA'")->Expand('Orders');

  2. Yii extension odataclient:

  $client = new EODataClient('northwind');
  $query = $client->createEntitySet('Customers')->filter("Country eq 'USA'")->Expand('Orders');
     * @param string $name
     * @return object
     */
    public function createEntitySet($name)
    {
        if (method_exists($this->service, $name)) {
            return $this->service->$name();
        }else
            throw new CHttpException(400, "EntitySet '$name' is not defined in service " . $this->_serviceName);
    }

    /**
     * Get a unique id of the current service for memory caching
     *
     * @return string
     */
    public function getServiceId()
    {
        return md5($this->_serviceName . serialize($this->_config));
    }

    /**
     * Get the config for the current service
     *
     * @return
     */
    public function getConfig()
    {
        return $this->_config;
    }

    /**
     * Get the configured serviceUri from the config file
     *
     * @return string
     */
    public function getServiceUri()
    {
        return $this->_config['serviceUri'];
    }

    /**
     * Get the configured serviceSubDir from the config file
     *
     * @return string
     */
    public function getServiceSubDir()
    {
        return isset($this->_config['serviceSubDir']) ? $this->_config['serviceSubDir'] : '';
    }

    /**
     * Get the configured options from the config file
     *
     * @return array
     */
    public function getOptions()
    {
        return $this->_config['options'];
    }

    /**
     * Get the real path to the classes directory
     *
     * @param boolean $checkWriteable
     * @return string
     */
    public static function getServiceClassesPath($checkWriteable = false)
    {
        $path = self::getConfigPath() . '/classes';

        if ($checkWriteable && !is_writable($path))
            throw new CException('Service classes directory is not writeable.');

        return $path;
    }

    /**
     * Get the real path to the service classes directory
     * Checks if writeable
     * Add a subDir if set in config file
     * Create the dir/subdir with the specified permissions if not exists
     *
     * @param boolean $checkWriteable
     * @param integer $permissions
     * @return string
     */
    public function getServiceClassPath($checkWriteable = false, $permissions = 0754)
    {
        $path = self::getServiceClassesPath($checkWriteable);

        if (!empty($this->serviceSubDir))
            $path .= '/' . $this->serviceSubDir;

        if (!is_dir($path)) {
            if (!@mkdir($path, $permissions, true))
                'Unable to create empty service class directory for service: ' . $this->_serviceName;
        }

        return $path;
    }

    /**
     * Get the path to the stored proxy class file
     *
     * @return string
     */
    public function getServicefile($checkWriteable = false, $permissions = 0754)
    {
        return $this->getServiceClassPath($checkWriteable, $permissions) . '/' . $this->_serviceName . '.php';
    }

    /**
     * Get the path to the stored metadata file
     *
     * @return string
     */
    public function getMetadataFile()
    {
        return $this->getServiceClassPath() . '/' . $this->_serviceName . '.xml';
    }

    /**
     * Include the service proxy class file
     */
    public function includeProxyClassFile()
    {
        $serviceFile = $this->getServicefile();

        if (!is_file($serviceFile))
            throw new CException('Service class file not found for service: ' . $this->_serviceName);

        require_once($serviceFile);
    }

    /**
     * Add the SDK to the include_path
     */
    public static function importSDK()
    {
        $odataSdk = Yii::getPathOfAlias('ext.odataclient.vendors.OData_PHP_SDK.framework');
        $includePath = get_include_path();

        if (strpos($includePath, $odataSdk) === false)
            set_include_path($includePath . PATH_SEPARATOR . $odataSdk);
    }

    /**
     * Add the SDK to include_path and include the services proxy class
     */
    public function importService()
    {
        if (!$this->_serviceImported) {
            self::importSDK();
            $this->includeProxyClassFile();
            $this->_serviceImported = true;
        }
    }

	/**
	 * Add the SDK to include_path and include the services proxy class
	 *
	 * @param string $serviceProxyClass
	 */
	public static function importServices($serviceProxyClass)
	{
		self::importSDK();
	    Yii::import($serviceProxyClass);
	}

    /**
     * Generate a service class definition file
     * Uses the (modified) PHPDataSvcUtil.php from SDK to download the class definitions
     * New in PHPDataSvcUtil.php: save the metadatafile too
     */
    public function generateProxyClassFile()
    {
        // prepare the options as argv
        $argv = array('EOdataClient.php'); //Scriptname as first item, will be removed in PHPSvcUtil

        $options = array_merge($this->options,
            array(
                '/uri' => $this->serviceUri,
                '/out' => $this->getServicefile(true),
                ));

        foreach ($options as $key => $value)
        $argv[] = "$key=$value";

        self::importSDK();
        require_once('PHPDataSvcUtil.php');
    }

    /**
     * Load the metadatafile into a SimpleXMLElement object
     *
     * @return SimpleXMLElement
     */
    public function getMetadataXmlElememt()
    {
        if (!isset(self::$_metadataXmlElement))
            self::$_metadataXmlElement = array();

        $id = $this->getServiceId();

        if (!isset(self::$_metadataXmlElement[$id])) {
            $metadataFile = $this->getMetadataFile();
            if (!is_file($metadataFile))
                throw new CException('Metadata file not found for service: ' . $this->_serviceName);
            // do str_replace because of problems with Namespaces and XPath
            // @link http://php.net/manual/en/simplexmlelement.xpath.php
            $xmlStr = str_replace('xmlns=', 'ns=', file_get_contents($metadataFile));

            self::$_metadataXmlElement[$id] = new SimpleXMLElement($xmlStr);
        }
        return self::$_metadataXmlElement[$id];
    }

    /**
     * Get the corresponding EntityTypeName for the specified $entitySetName
     *
     * @param string $entitySetName
     * @return string
     */
    public function getEntityTypeName($entitySetName)
    {
        $entities = $this->getEntities();

        if (!array_key_exists($entitySetName, $entities))
            throw new CException('Entityset not defined in metadata: ' . $entitySetName);

        return $entities[$entitySetName];
    }

    /**
     * Extract the keys of a data object from the metadatafile
     * Used in the EODataClientDataprovider.fetchKeys()
     *
     * @param string $entitySetName
     * @return array
     */
    public function getEntityKeys($entitySetName)
    {
        $result = array();

        $entityTypeName = $this->getEntityTypeName($entitySetName);

        $xml = $this->getMetadataXmlElememt();
        $xpath = "/edmx:Edmx/edmx:DataServices/Schema/EntityType[@Name=\"$entityTypeName\"]/Key";

        $xpathResult = $xml->xpath($xpath);
        foreach ($xpathResult as $node)
        $result[] = (string)$node->PropertyRef['Name'];

        return $result;
    }

    /**
     * Extract the attributes (=Property) of the specified $entitySetName from the metadatafile
     *
     * @param string $entitySetName
     * @return array
     */
    public function getEntityAttributes($entitySetName)
    {
        $result = array();

        $entityTypeName = $this->getEntityTypeName($entitySetName);

        $xml = $this->getMetadataXmlElememt();
        $xpath = "/edmx:Edmx/edmx:DataServices/Schema/EntityType[@Name=\"$entityTypeName\"]/Property";

        $xpathResult = $xml->xpath($xpath);
        foreach ($xpathResult as $node)
        $result[(string)$node['Name']] = $node;

        return $result;
    }

    /**
     * Get an array of entitySets with Name as key and EntityType as value
     *
     * @return array
     */
    public function getEntities()
    {
        if (!isset(self::$_entities))
            self::$_entities = array();

        $id = $this->getServiceId();

        if (!isset(self::$_entities[$id])) {
            $result = array();

            $xml = $this->getMetadataXmlElememt();
            $xpath = "//EntityContainer/EntitySet";
            $xpathResult = $xml->xpath($xpath);

            foreach ($xpathResult as $node)
            $result[(string)$node['Name']] = substr(strrchr((string)$node['EntityType'], '.'), 1);

            self::$_entities[$id] = $result;
        }

        return self::$_entities[$id];
    }

    /**
     * Get a dataProvider for rendering a list of entitySets
     *
     * @see ODataController.actionServices
     * @param array $config the dataProvider config
     * @return CArrayDataProvider
     */
    public function getEntitiesDataProvider($config = array())
    {
        $entities = $this->getEntities();
        $rawData = array();
        foreach ($entities as $key => $value)
        $rawData[] = array('entitySet' => $key, 'entityType' => $value);

        $config = array_merge($config, array('keyField' => 'entitySet'));
        return new CArrayDataProvider($rawData, $config);
    }

    /**
     * Get a dataProvider with data from the config file
     *
     * @see ODataController.actionIndex
     * @param string $configFileName
     * @param array $config
     * @return
     */
    public static function getConfiguredServicesDataProvider($configFileName = 'config.php', $config = array())
    {
        $cfg = new CConfiguration(self::getConfigFile($configFileName));
        $configServices = $cfg->toArray();

        $rawData = array();
        foreach ($configServices as $key => $value)
        $rawData[] = array('service' => $key, 'config' => $value);

        $config = array_merge($config, array('keyField' => 'service'));
        return new CArrayDataProvider($rawData, $config);
    }

    /**
     * Execute a query, catching exceptions
     *
     * @param object $query
     * @return Response
     */
    public function execQuery($query)
    {
        try {
            return $query->Execute();
        }
        catch(DataServiceRequestException $e) {
            throw new CException($e->Response->getError() . "\n" . $e->Response->getQuery());
        }
    }
}