<?php

/**
 * EODataClientDataProvider.php
 *
 * A DataProvider for data loaded from OData services
 *
 * @link http://www.odata.org/
 *
 * @author Joe Blocher <yii@myticket.at>
 * @copyright 2011 myticket it-solutions gmbh
 * @license New BSD License
 * @category Webservice
 * @package odataclient
 * @version 0.1
 */
Yii::import('ext.odataclient.components.EODataClient');

/**
 * EODataClientDataProvider
 *
 * @package
 * @author joe
 * @copyright Copyright (c) 2011
 * @version $Id$
 * @access public
 */
class EODataClientDataProvider extends CDataProvider
{
	/**
	 * The EODataClient instance
	 *
	 * @var EODataClient $_client
	 */
	private $_client;

	/**
	 * The name of the entitySet
	 *
	 * @var string $_entitySetName
	 */
	private $_entitySetName;

	/**
	 * The DataServiceQuery object from the SDK
	 *
	 * @var DataServiceQuery $_query
	 */
	private $_query;

	/**
	 * The Response object from the SDK
	 *
	 * @var QueryOperationResponse $_response
	 */
	private $_response;

	/**
	 * Create the DataProvider
	 * @see ODataController.actionServices
	 *
	 * @param object $client the EODataClient instance
	 * @param string $entitySetName the name of the entityset ('Customers', 'Categories' ...)
	 * @param DataServiceQuery $query the data query: if null all data are loaded from the EntitySet
	 * @param array $config the config for the dataprovider (pagination, sort ...)
	 */
	public function __construct($client,$entitySetName,$query=null,$config = array())
	{
		$this->_client=$client;
		$this->_entitySetName=$entitySetName;
		$this->_client->importService();

		$this->_query = isset($query) ? $query : $client->createEntitySet($entitySetName);

		foreach($config as $key=>$value)
			$this->$key=$value;
	}

	/**
	 * Get the ResponseObject
	 *
	 * @return QueryOperationResponse
	 */
	public function getResponse()
	{
	   	return $this->_response;
	}

	/**
	 * Get the EODataClient instance
	 *
	 * @return EODataClient
	 */
	public function getClient()
	{
		return $this->_client;
	}

	/**
	 * Get the entitySetName
	 *
	 * @return string
	 */
	public function getEntitySetName()
	{
		return $this->_entitySetName;
	}

	/**
	 * Get the DataServiceQuery object
	 *
	 * @return DataServiceQuery
	 */
	public function getQuery()
	{
		return $this->_query;
	}

	/**
	 * Fetches the data from the persistent data storage.
	 * @return array list of data items
	 */
	protected function fetchData()
	{
		$result = array();

		if(($pagination=$this->getPagination())!==false)
		{
			$pagination->setItemCount($this->getTotalItemCount());
			$this->_query->Skip($pagination->getOffset());
			$this->_query->Top($pagination->getLimit());
		}

		if(($sort=$this->getSort())!==false)
		{
			$order=$sort->getOrderBy();

			if (empty($order) && !empty($sort->attributes))
				$order = implode(',',$sort->attributes);

			//need lowercase
			$order = str_ireplace(' ASC',' asc',$order);
			$order = str_ireplace(' DESC',' desc',$order);

			if (!empty($order))
				$this->_query->OrderBy($order);
		}

		$this->_response = $this->_client->execQuery($this->_query);

		foreach($this->_response->Result as $item)
			$result[] = $item;

	    return $result;
	}

	/**
	 * Fetches the data item keys from the persistent data storage.
	 * @return array list of data item keys.
	 */
	protected function fetchKeys()
	{
		return $this->_client->getEntityKeys($this->_entitySetName);
	}

	/**
	 * Calculates the total number of data items.
	 * @return integer the total number of data items.
	 */
	protected function calculateTotalItemCount()
	{
		try
		{
		  return $this->_query->Count();
		}
		catch(DataServiceRequestException $e)
		{
			throw new CException($e->Response->getError()."\n".$e->Response->getQuery());
		}
	}
}